var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
import axios from 'axios';
import fsp from 'node:fs/promises';
import fs from 'node:fs';
import { strict as assert } from 'assert';
import PromisePool from '@supercharge/promise-pool';
import * as http from 'http';
import * as https from 'https';
import EventEmitter from 'events';
const DEFAULT_CHUNK_SIZE = 16 * 1024 * 1024;
const MIN_CHUNK_SIZE = 1024;
const DEFAULT_CONCURRENCY = 4;
const DEFAULT_RETRY_COUNT = 10;
export default class TurboDownloader extends EventEmitter {
    constructor(options) {
        var _a;
        super();
        this.started = false;
        this.aborted = false;
        this.abortSaveProgress = false;
        this.abortHandlers = [];
        this.httpAgent = new http.Agent({
            keepAlive: true,
            maxSockets: 1024,
            timeout: 30000,
        });
        this.httpsAgent = new https.Agent({
            keepAlive: true,
            maxSockets: 1024,
            timeout: 30000,
        });
        this.options = Object.assign(Object.assign({}, options), { chunkSize: options.chunkSize || DEFAULT_CHUNK_SIZE, concurrency: options.concurrency || DEFAULT_CONCURRENCY, retryCount: options.retryCount || DEFAULT_RETRY_COUNT, canBeResumed: (_a = options.canBeResumed) !== null && _a !== void 0 ? _a : true, adapter: options.adapter, fillFileByte: options.fillFileByte || 0, transformStream: options.transformStream });
        assert(this.options.chunkSize >= MIN_CHUNK_SIZE, `Chunk size must be ${MIN_CHUNK_SIZE} or greater`);
        assert(this.options.concurrency >= 1, 'Concurrency must be 1 or greater');
        assert(this.options.retryCount >= 0, 'retryCount must be 0 or greater');
    }
    download(progressCallback) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.started) {
                throw new Error('Already started');
            }
            this.started = true;
            this.emit('downloadStarted', this.options.url, this.options.destFile);
            const plan = yield this.getDownloadingPlan();
            this.emit('planReady', plan);
            const chunksToDownload = plan.chunks.filter((chunk) => chunk.downloaded < chunk.size);
            try {
                yield PromisePool.for(chunksToDownload)
                    .withConcurrency(this.options.concurrency)
                    .handleError((error) => __awaiter(this, void 0, void 0, function* () {
                    throw error;
                }))
                    .process((chunk) => __awaiter(this, void 0, void 0, function* () {
                    if (this.aborted) {
                        return;
                    }
                    let retries = 0;
                    let lastError;
                    while (retries <= this.options.retryCount) {
                        try {
                            this.emit('chunkDownloadStarted', chunk, retries);
                            yield this.downloadChunk(chunk, (newChunk) => {
                                chunk.downloaded = newChunk.downloaded;
                                if (this.options.canBeResumed) {
                                    this.saveDownloadingPlanToDisk(plan);
                                }
                                if (plan.size > 0 && progressCallback) {
                                    const downloaded = plan.chunks.reduce((sum, chunk) => sum + chunk.downloaded, 0);
                                    progressCallback(downloaded, plan.size, plan);
                                }
                            }, (abort) => this.abortHandlers.push(abort));
                            this.emit('chunkDownloadFinished', chunk, retries);
                            return;
                        }
                        catch (e) {
                            this.emit('chunkDownloadError', chunk, retries, e);
                            lastError = e;
                            retries++;
                            yield new Promise((resolve) => {
                                setTimeout(resolve, 1000 * Math.pow(retries, 2));
                            });
                        }
                    }
                    throw lastError;
                }));
            }
            catch (e) {
                this.emit('downloadError', this.options.url, this.options.destFile);
                if (!this.options.canBeResumed) {
                    yield this.deletePlanFromDisk();
                }
                throw e;
            }
            if (!this.abortSaveProgress) {
                yield this.deletePlanFromDisk();
                if (this.aborted) {
                    yield fsp.unlink(this.options.destFile);
                }
            }
        });
    }
    abort(saveProgress = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.started) {
                return;
            }
            this.aborted = true;
            this.emit('aborted');
            this.abortSaveProgress = saveProgress;
            for (const handler of this.abortHandlers) {
                handler();
            }
        });
    }
    deletePlanFromDisk() {
        return __awaiter(this, void 0, void 0, function* () {
            const fileName = this.getPlanFileName();
            if (fs.existsSync(fileName)) {
                yield fsp.unlink(fileName);
            }
        });
    }
    downloadChunk(chunk, progressCallback, abortHandler) {
        return __awaiter(this, void 0, void 0, function* () {
            const start = chunk.size > 0 ? chunk.disposition + chunk.downloaded : 0;
            const sizeLeft = chunk.size > 0 ? chunk.size - chunk.downloaded : -1;
            const cancelTokenSource = axios.CancelToken.source();
            const options = {
                httpAgent: this.httpAgent,
                httpsAgent: this.httpsAgent,
                responseType: 'stream',
                cancelToken: cancelTokenSource.token,
                adapter: this.options.adapter,
            };
            if (sizeLeft > 0) {
                options.headers = { range: `bytes=${start}-${start + sizeLeft - 1}` };
            }
            const response = yield axios.get(this.options.url, options);
            const responseStream = response.data;
            const stream = this.options.transformStream
                ? this.options.transformStream(responseStream)
                : responseStream;
            const fd = yield fsp.open(this.options.destFile, 'r+');
            const fileStream = fd.createWriteStream({ start: chunk.disposition });
            yield new Promise((resolve, reject) => {
                abortHandler(() => {
                    cancelTokenSource.cancel();
                });
                stream.on('data', (buffer) => __awaiter(this, void 0, void 0, function* () {
                    chunk.downloaded += buffer.length;
                    this.emit('chunkDownloadProgress', chunk);
                    progressCallback(chunk);
                }));
                stream.on('error', (err) => {
                    reject(err);
                });
                stream.pipe(fileStream);
                fileStream.on('close', () => {
                    resolve();
                });
            });
        });
    }
    getDownloadingPlan() {
        return __awaiter(this, void 0, void 0, function* () {
            const options = yield this.getOptions();
            return ((yield this.readDownloadingPlanFromDisk(options)) ||
                (yield this.createDownloadingPlan(options)));
        });
    }
    getPlanFileName() {
        return `${this.options.destFile}.turbodownload`;
    }
    readDownloadingPlanFromDisk(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const fileName = this.getPlanFileName();
            if (fs.existsSync(fileName)) {
                const data = yield fsp.readFile(this.getPlanFileName(), 'utf8');
                try {
                    const plan = JSON.parse(data);
                    if (plan.acceptRanges === options.acceptRanges &&
                        plan.size === options.size) {
                        return plan;
                    }
                }
                catch (e) { }
            }
            return null;
        });
    }
    saveDownloadingPlanToDisk(plan) {
        fs.writeFileSync(this.getPlanFileName(), JSON.stringify(plan, null, 2), 'utf8');
    }
    createDownloadingPlan(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const plan = Object.assign(Object.assign({}, options), { chunks: [] });
            if (plan.acceptRanges && plan.size && plan.size > 0) {
                const chunkSize = this.options.chunkSize;
                const chunksCount = Math.ceil(plan.size / chunkSize);
                for (let i = 0; i < chunksCount; i++) {
                    const size = i === chunksCount - 1 ? plan.size - chunkSize * i : chunkSize;
                    plan.chunks.push({
                        size,
                        disposition: i * chunkSize,
                        downloaded: 0,
                    });
                }
            }
            else {
                plan.chunks.push({
                    size: plan.size,
                    disposition: 0,
                    downloaded: 0,
                });
            }
            this.emit('reservingSpaceStarted', options.size);
            yield this.reserveSpace(options);
            this.emit('reservingSpaceFinished', options.size);
            return plan;
        });
    }
    getOptions() {
        return __awaiter(this, void 0, void 0, function* () {
            const reqOptions = {
                httpAgent: this.httpAgent,
                httpsAgent: this.httpsAgent,
                adapter: this.options.adapter,
            };
            const response = yield axios.head(this.options.url, reqOptions);
            const acceptRanges = response.headers['accept-ranges'] === 'bytes';
            const size = response.headers['content-length']
                ? parseInt(response.headers['content-length'])
                : -1;
            return {
                acceptRanges,
                size,
            };
        });
    }
    reserveSpace(options) {
        return __awaiter(this, void 0, void 0, function* () {
            const fd = yield fsp.open(this.options.destFile, 'w');
            if (options.size > 0) {
                const buffer = Buffer.alloc(DEFAULT_CHUNK_SIZE).fill(this.options.fillFileByte);
                let wrote = 0;
                while (wrote < options.size) {
                    const sz = Math.min(buffer.length, options.size - wrote);
                    if (sz > 0) {
                        yield fd.write(buffer, 0, sz);
                        wrote += sz;
                    }
                }
            }
            yield fd.close();
        });
    }
}
